@@ -7,6 +7,7 @@ module Exn        = Exn
 module Filename   = Filename
 module Hashtbl    = Hashtbl
 module Int        = Int
+module Id         = Id
 module Io         = Io
 module List       = List
 module Map        = Map
@@ -31,6 +32,13 @@ module Env        = Env
 module Proc       = Proc
 module Type_eq    = Type_eq
 module Nothing    = Nothing
+module Bin        = Bin
+module Digest     = Digest
+module Fdecl      = Fdecl
+module Unit       = Unit
+module Monad      = Monad
+module Fn         = Fn
+module Dyn        = Dyn
 
 external reraise : exn -> _ = "%reraise"
 
